/*
  Copyright (C) 1995-2023 Free Software Foundation, Inc.

  This file is part of GNU Inetutils.

  GNU Inetutils is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or (at
  your option) any later version.

  GNU Inetutils is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see `http://www.gnu.org/licenses/'. */

/*
 * Copyright (c) 1988, 1993
 *	The Regents of the University of California.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <config.h>

#include <stdlib.h>

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <time.h>
#include <errno.h>

#include <arpa/telnet.h>
#include <sys/select.h>

#include "ring.h"

#include "defines.h"
#include "externs.h"

Ring netoring, netiring;
unsigned char netobuf[2 * BUFSIZ], netibuf[BUFSIZ];

/*
 * Initialize internal network data structures.
 */

void
init_network (void)
{
  if (ring_init (&netoring, netobuf, sizeof netobuf) != 1)
    {
      exit (EXIT_FAILURE);
    }
  if (ring_init (&netiring, netibuf, sizeof netibuf) != 1)
    {
      exit (EXIT_FAILURE);
    }
  NetTrace = stdout;
}


/*
 * Check to see if any out-of-band data exists on a socket (for
 * Telnet "synch" processing).
 */

int
stilloob (void)
{
  static struct timeval timeout = { 0, 0 };
  fd_set excepts;
  int value;

  do
    {
      FD_ZERO (&excepts);
      FD_SET (net, &excepts);
      value =
	select (net + 1, (fd_set *) 0, (fd_set *) 0, &excepts, &timeout);
    }
  while ((value == -1) && (errno == EINTR));

  if (value < 0)
    {
      perror ("select");
      (void) quit (0, NULL);
      /* NOTREACHED */
    }
  if (FD_ISSET (net, &excepts))
    {
      return 1;
    }
  else
    {
      return 0;
    }
}


/*
 *  setneturg()
 *
 *	Sets "neturg" to the current location.
 */

void
setneturg (void)
{
  ring_mark (&netoring);
}


/*
 *  netflush
 *		Send as much data as possible to the network,
 *	handling requests for urgent data.
 *
 *		The return value indicates whether we did any
 *	useful work.
 */


int
netflush (void)
{
  register int n, n1;

#ifdef	ENCRYPTION
  if (encrypt_output)
    ring_encrypt (&netoring, encrypt_output);
#endif /* ENCRYPTION */
  if ((n1 = n = ring_full_consecutive (&netoring)) > 0)
    {
      if (!ring_at_mark (&netoring))
	{
	  n = send (net, (char *) netoring.consume, n, 0);	/* normal write */
	}
      else
	{
	  /*
	   * In 4.2 (and 4.3) systems, there is some question about
	   * what byte in a sendOOB operation is the "OOB" data.
	   * To make ourselves compatible, we only send ONE byte
	   * out of band, the one WE THINK should be OOB (though
	   * we really have more the TCP philosophy of urgent data
	   * rather than the Unix philosophy of OOB data).
	   */
	  n = send (net, (char *) netoring.consume, 1, MSG_OOB);	/* URGENT data */
	}
    }
  if (n < 0)
    {
      if (errno != ENOBUFS && errno != EWOULDBLOCK)
	{
	  setcommandmode ();
	  perror (hostname);
	  (void) NetClose (net);
	  ring_clear_mark (&netoring);
	  longjmp (peerdied, -1);
	}
      n = 0;
    }
  if (netdata && n)
    {
      Dump ('>', netoring.consume, n);
    }
  if (n)
    {
      ring_consumed (&netoring, n);
      /*
       * If we sent all, and more to send, then recurse to pick
       * up the other half.
       */
      if ((n1 == n) && ring_full_consecutive (&netoring))
	{
	  (void) netflush ();
	}
      return 1;
    }
  else
    {
      return 0;
    }
}
